/* $Workfile: A66BurstEx.c $       $Revision: 9 $      $Modtime: 9/12/05 5:49p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    A66BurstEx.c

    A simple example console application demonstrating the use of the
    DMA Direct Sending and Receiving with the OTX Driver using the
    board's Burst device to access the data using DMA.

      Arni-6x6-PCI-Plus
      Arni-6x6-PCI

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 1 DSP that sends and receives data via Burst.

    The direct sends and receives operate on a highway basis and are
    hosted by the Burst device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#define DEMO_DISPLAY_BYTES 1   // Comments this macro out to display only the received F-bits (T1 mode only) in verbose mode

#define DEMO_DEFAULT_LOG_FILE   "A66BurstEx.log"
#define DEMO_DEFAULT_BOARD_NO   0

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k


typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int             coDsp;
  int             boardNo;
  int             nCurrentDsp;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE        *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;


struct {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hDsp[8];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;

struct {
  OTX_HANDLE hDspIn[8];
  OTX_HANDLE hDspOut[8];
} g_hPipes;

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)
};

struct {
  struct sBurstSender mBurstSender;        // This board has one highway for the burst device
  struct sBurstReceiver mBurstReceiver;    // This board has one highway for the burst device
} g_sBurstDevs;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Disable the Watchdog Timer, and enable it when we have completed the configuration
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Get a handle to the Time-Space Switch (TSS) on the board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    DemoLogf("Open OTX_DEVICE_DSP\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
#if 0
    DemoLogf("Loading OTX_DSP_SPM_OTXCST2 to DSP#%d \n", i);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXCST2); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#else
    DemoLogf("Loading OTX_DSP_SPM_OTXSPM2 to DSP#%d \n", i);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXSPM2); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#endif
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
}




void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hBoardController0\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    // Enable Physical Devices, Dsp's
    DemoLogf("Enable g_hPhysDevs.hDsp[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
#if 0
  OTX_INT32 nDsp;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[32];


  // Create the Pipes (one for each direction) between the Burst Device and the DSPs
  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    OTX_INT32 nTs;

    // Burst highway number [0]
    iHighwayNo = 0;

    // Create pipes (one for each direction) Use all timeslots except TS0.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    DemoLogf("Creating pipe for Burst Sender/Receiver\n");

    // Connect only one timeslot for now
    nTs = 0;
      
    sqTimeSlotMask[nTs] = 0xFF;
    sqBurstTimeSlotMask[nTs] = 0xFF;


    // --------------------------------------------------------------------
    // Create an incoming and outgoing Pipe 
    // --------------------------------------------------------------------
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspIn[nDsp]);
    DemoCheckRc(g_hPipes.hDspIn[nDsp], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspOut[nDsp]);
    DemoCheckRc(g_hPipes.hDspOut[nDsp], result);

    // -------------------------
    // Connect pipe: Burst -> Dsp
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hDspIn[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hDspIn[nDsp], result);

    result = OtxDrvConnectPipe(g_hPipes.hDspIn[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hDspIn[nDsp], result);

    // -------------------------
    // Connect pipe: Dsp -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hDspOut[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hDspOut[nDsp], result);

    result = OtxDrvConnectPipe(g_hPipes.hDspOut[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hDspOut[nDsp], result);

  }
#endif
#if 1
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo = 0;  // Internal Burst highway number [0]
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_HANDLE hPipe;
  OTX_INT32 nTs;
  
  
  DemoLogf("Creating pipe for Burst Sender/Receiver. Looped through the TSS device\n");
  
  // Connect only one timeslot for now
  nTs = 0;
  memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
  sqTimeSlotMask[nTs] = 0xFF;
  
  
  // --------------------------------------------------------------------
  // Create a Pipe 
  // --------------------------------------------------------------------
  result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(hPipe, result);
  
  // -------------------------
  // Connect pipe: Burst -> Burst
  // -------------------------
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqTimeSlotMask, 32);
  DemoCheckRc(hPipe, result);
  
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqTimeSlotMask, 32);
  DemoCheckRc(hPipe, result);

#endif

}



/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX [%d.%d])\n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result));
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif

    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
#elif defined(WIN32)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
#endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              continue;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              DemoBurstSender(eventData.m_nRequestId);
              continue;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.
*******************************************************************************/
void *DemoStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT        result;
  OTX_DATETIME      nTimeStamp;
  OTX_UINT32        stat_Starts;            // Number of actual dma transfers initiated
  OTX_UINT32        stat_Completes;         // Number of actual dma transfers completed
  OTX_UINT32        stat_Partial;           // Number of partial transfer due to framemask
  OTX_UINT32        stat_Full;              // Number of count of max dma size transfers
  OTX_UINT32        stat_Aborts;            // Number of actual dma transfers aborts
  OTX_UINT32        stat_Amount;            // Number of amount of data transfered
  OTX_UINT32        stat_Errors;            // Number of count of processing errors
  OTX_UINT32        stat_OUruns;            // Number of amount of over/under runs
  OTX_UINT32        stat_OUrunsUser;        // Number of amount of over/under runs caused by user mode
  OTX_UINT32        stat_DmaBusy;           // Number of count of DMA busy occurances
  OTX_UINT32        stat_Notify;            // Number of logical device notifies
  OTX_UINT32        stat_NotifyPartial;     // Number of logical device notifies - partial
  OTX_UINT32        stat_FifoSize;          // size of the Burst FIFO
  OTX_UINT32        stat_FifoTrigger;       // the current FIFO Trigger value
  OTX_UINT32        stat_FifoCount;         // the current amount in the FIFO
  OTX_UINT32        stat_FifoPeak;          // the peak FIFO usage
  OTX_UINT32        stat_FifoTriggerCount;  // the current FIFO trigger fire value

#define STATS_SHOW_FRAME_REF
#ifdef STATS_SHOW_FRAME_REF
  OTX_UINT32                ui32;
  OtxBurstFrameRefAndTime   FrameRef;             // frame reference data
  static OTX_UINT64         nLastFrame = 0;
#endif  // #ifdef STATS_SHOW_FRAME_REF
  

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);
    
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS,             &stat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES,          &stat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL,            &stat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL,               &stat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS,             &stat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT,             &stat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS,             &stat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS,             &stat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER,        &stat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY,            &stat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY,             &stat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL,     &stat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE,           &stat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,        &stat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_COUNT,          &stat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_PEAK,           &stat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER_COUNT,  &stat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d, FIFO Size:%d, Trigger:%d, Count:%d, Peak:%d, TriggerCount:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial,
        stat_FifoSize,
        stat_FifoTrigger,
        stat_FifoCount,
        stat_FifoPeak,
        stat_FifoTriggerCount
    );

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS,             &stat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES,          &stat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL,            &stat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL,               &stat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS,             &stat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT,             &stat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS,             &stat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS,             &stat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER,        &stat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY,            &stat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY,             &stat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL,     &stat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_SIZE,             &stat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,          &stat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_COUNT,            &stat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_PEAK,             &stat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER_COUNT,    &stat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d, FIFO Size:%d, Trigger:%d, Count:%d, Peak:%d, TriggerCount:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial,
        stat_FifoSize,
        stat_FifoTrigger,
        stat_FifoCount,
        stat_FifoPeak,
        stat_FifoTriggerCount
    );

#ifdef STATS_SHOW_FRAME_REF
    result = OtxDrvGetAttributeValue(g_hPhysDevs.hBurst, OTX_ATTR_BURST_GET_FRAME_REF_DATA, &FrameRef, sizeof(FrameRef), &ui32); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // just skip if we wrapped
    if (FrameRef.m_FrameCounter > nLastFrame) {
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("Burst FRef:: %d fps  Cntr:%08x%08x  Time:%08x%08x ( ",
        nLastFrame == 0 ? 0 : (OTX_UINT32) (FrameRef.m_FrameCounter - nLastFrame),
        (OTX_UINT32) (FrameRef.m_FrameCounter >> 32),
        (OTX_UINT32) (FrameRef.m_FrameCounter),
        FrameRef.m_FrameDateTime.sUint32.nHigh,
        FrameRef.m_FrameDateTime.sUint32.nLow
      );
      DemoPrintTimeStamp(FrameRef.m_FrameDateTime);
      DemoLogf(")\n");
    }
    nLastFrame = FrameRef.m_FrameCounter;
#endif  // #ifdef STATS_SHOW_FRAME_REF

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

  if (g_sBurstDevs.mBurstSender.pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], 0xFF, sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]));

    if (g_sBurstDevs.mBurstSender.nTsToSend == -1) {
      // All TS
      g_sBurstDevs.mBurstSender.cbTxBuf = fread(
        &g_sBurstDevs.mBurstSender.sqTxBuf,
        sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]),
        sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]),
        g_sBurstDevs.mBurstSender.pTxFile
      );

    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32];
      int cbSingleTsBuf;
      int byteCount = 0;
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        sizeof(singleTsBuf) / sizeof(singleTsBuf[0]),
        g_sBurstDevs.mBurstSender.pTxFile
        );

      
      // Dispurse the single TS to be transmitted in the whole transmit buffer 
      byteCount = 0;
      for ( x = g_sBurstDevs.mBurstSender.nTsToSend; 
      (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
      x += 32
        ) {
        g_sBurstDevs.mBurstSender.sqTxBuf[x] = singleTsBuf[byteCount++];
      }
      
      assert(byteCount == cbSingleTsBuf);
      g_sBurstDevs.mBurstSender.cbTxBuf = cbSingleTsBuf * 32;

      if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
        assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
      } else {
       DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
      }

    }
        
    if (g_sBurstDevs.mBurstSender.cbTxBuf > 0) {

      DemoLogf("DemoBurstSender[%d]: Transmitting %d bytes\n", nSourceId, g_sBurstDevs.mBurstSender.cbTxBuf);

      ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);

    } else {

      if (g_sBurstDevs.mBurstSender.bFileLoop) {
        if (g_sBurstDevs.mBurstSender.bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_sBurstDevs.mBurstSender.pTxFile);
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_sBurstDevs.mBurstSender.pTxFile);
        g_sBurstDevs.mBurstSender.pTxFile = NULL;
        g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;
      }
    }

  } else {
    // Send a text string (continously until aborted)

    // we MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)

    g_sBurstDevs.mBurstSender.cbTxBuf = g_sBurstDevs.mBurstSender.cbTxBuf;
    assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);

    ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_sBurstDevs.mBurstReceiver.bVerbose) {
    int x;

#ifdef DEMO_DISPLAY_BYTES // Display the first 32 bytes of from the current Dsp
    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d] (len=%.5d) ID:%08x ", nSourceId, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 64) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if ((x % 8) == 0) DemoLogf("\n");
      DemoLogf("%.2X ", buf[x]);
    }
    DemoLogf("\n");
    DemoLogf("\n");
#else 
    xx
#endif
  }

  // write data to a file if desired
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    int cbWritten;

    if (g_sBurstDevs.mBurstReceiver.nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      for (
          x = g_sBurstDevs.mBurstReceiver.nTsToRecord;
          x < (OTX_INT32) sizeof(buf); 
          x += 32
      ) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}


/*******************************************************************************

  @func

*******************************************************************************/
OTX_BOOL DemoIsBurstDeviceEnabled()
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  result = OtxDrvGetState(g_hPhysDevs.hBurst, &eDevState); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The burst device is enabled
  } else {
    return OTX_FALSE; // The burst device is not enabled
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select active Dsp or Dsp, this Dsp/Dsp is then used for subsequent commands\n");
    DemoLogf(" e   Enable Burst\n");
    DemoLogf(" d   Disable Burst\n");
    DemoLogf(" a   Abort Burst[%d] Senders and Receivers\n",  0);
    DemoLogf(" r   Start Burst Receiver[%d] (no file logging)\n", 0);
    DemoLogf(" R   Start Burst Receiver[%d] Prompt for file to recv\n", 0);
    DemoLogf(" s   Start Burst Sender[%d]. Continously send a text string\n", 0);
    DemoLogf(" S   Start Burst Sender[%d]. Prompt for file to send\n", 0);
    DemoLogf(" L   Burst Sender[%d]. Toggle loop file sends (currently %s)\n", 0, g_sBurstDevs.mBurstSender.bFileLoop ? "On" : "Off");
    DemoLogf(" v   Toggle verbose (print received data) setting Burst Receiver[%d]\n", 0);
    DemoLogf(" V   Toggle verbose (print send progress) setting Burst Sender[%d]\n", 0);
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" t   Toggle verbose buffer usage statistics\n");
    DemoLogf(" f   Forcing resynchronization\n");
    DemoLogf(" F   Reset all lines\n");
    DemoLogf(" T   Reset buffer usage statistics\n");
    DemoLogf(" D   Dump buffer usage statistics\n");
    DemoLogf(" n   Set burst notification value (1-32)\n");
    break;

  case 'e':
    // by enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing to connected logical devices.
    DemoLogf("Enable g_hPhysDevs.hBurst\n");
    result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 'd':
    // by disabling the Burst device, we stop the process of transferring data
    // from the hardware to the driver via DMA.
    DemoLogf("Disable g_hPhysDevs.hBurst\n");
    result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 't':
    g_bStatsThreadActive = !g_bStatsThreadActive;
    if (g_bStatsThreadActive) {
      DemoLogf("Starting Stats Thread\n");
      OsaThreadStart(DemoStatsThread);
    }
    break;

  case 'v':
    {
      if (DemoIsBurstDeviceEnabled()) {
        g_sBurstDevs.mBurstReceiver.bVerbose = !g_sBurstDevs.mBurstReceiver.bVerbose;
        DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_sBurstDevs.mBurstReceiver.bVerbose)?"ON":"OFF", 0);
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'V':
    {
      if (DemoIsBurstDeviceEnabled()) {
        g_sBurstDevs.mBurstSender.bVerbose = !g_sBurstDevs.mBurstSender.bVerbose;
        DemoLogf("Setting verbose setting to %s for Burst Sender[%d] Thread\n", (g_sBurstDevs.mBurstSender.bVerbose)?"ON":"OFF", 0);
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'a':
    {
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("DemoBurstReceiver: Closing file\n");
          fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
          g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
        }
        g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
        
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("mBurstSender: Closing file\n");
          fclose(g_sBurstDevs.mBurstSender.pTxFile);
          g_sBurstDevs.mBurstSender.pTxFile = NULL;
        }
        g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'r':
    {
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE) {
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        } else if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("DemoBurstReceiver: Closing file\n");
          fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
          g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
        } else {
          DemoLogf("Burst Receiver already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToRecord = -1;

      if (DemoIsBurstDeviceEnabled()) {
        nHighwayNo = 0;
        if (g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE || g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
          DemoLogf("BurstReceiver[%d]. Please enter file to receive: ", nHighwayNo);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            break;
          }
          
          if ((g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzBuf, "wb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          
          DemoLogf("BurstReceiver[%d]. Please enter single TS[0-31] to receive to file (-1=All) [%d]: ", nHighwayNo, nTsToRecord);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            break;
          } else {
            nTsToRecord = strtol(sqzBuf, NULL, 0);
          }
          g_sBurstDevs.mBurstReceiver.nTsToRecord = nTsToRecord;
          
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Receiver[%d] Thread already running.\n", nHighwayNo);
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 's':
    {
      OTX_UINT32 i;

      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstSender.bActive == OTX_FALSE) {
          DemoLogf("BurstSender. Continously send static values\n");
          
          g_sBurstDevs.mBurstSender.pTxFile = NULL;
          
          // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf).
          g_sBurstDevs.mBurstSender.cbTxBuf = 0;
          for (i = 0; i < DEMO_BURST_BUF_SIZE; i++) {
            
            g_sBurstDevs.mBurstSender.sqTxBuf[i] = (OTX_UINT8) i/32; // 32 zeros, 32 ones, etc
            
            g_sBurstDevs.mBurstSender.bActive = OTX_TRUE;
          }
          g_sBurstDevs.mBurstSender.cbTxBuf = DEMO_BURST_BUF_SIZE;
        } else {
          DemoLogf("Burst Sender already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToSend = -1;
      
      if (DemoIsBurstDeviceEnabled()) {
        nHighwayNo = 0;
        if (g_sBurstDevs.mBurstSender.bActive == OTX_FALSE) {
          DemoLogf("BurstSender[%d]. Please enter file to send: ", nHighwayNo);
          
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            break;
          }
          
          if ((g_sBurstDevs.mBurstSender.pTxFile = fopen(sqzBuf, "rb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          
          DemoLogf("BurstSender[%d]. Please enter single TS[0-%d] to send from file (-1=All) [%d]: ", nHighwayNo, nTsToSend);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0]) {
            nTsToSend = strtol(sqzBuf, NULL, 0);
          }
          g_sBurstDevs.mBurstSender.nTsToSend = nTsToSend;
          
          g_sBurstDevs.mBurstSender.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Sender[%d] already running.\n", nHighwayNo);
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'L':
    {
      g_sBurstDevs.mBurstSender.bFileLoop = !g_sBurstDevs.mBurstSender.bFileLoop;
      DemoLogf("Setting loop file sends %s for Burst Sender[%d] Thread\n", (g_sBurstDevs.mBurstSender.bFileLoop)?"ON":"OFF", 0);
    }
    break;

  case 'T':
    {
      // reset stats
      DemoLogf("Reseting statistics\n");
      OtxDrvDbgDump(g_hPhysDevs.hBurst, 101);
    }
    break;

  case 'D':
    {
      // dump stats to debugger
      DemoLogf("Dump stats to debugger\n");
      OtxDrvDbgDump(g_hPhysDevs.hBurst, 100);
    }
    break;

  case 'c':
    {
      OTX_UINT32 nCmd = 7; 
      static OTX_UINT8 nHighway = 0; 
      static OTX_UINT8 nTs = 0; 
      static OTX_UINT8 nVal = 0x55; 
      char strBuf[64];
      
      DemoLogf("TSS constByte\n ", nVal);
      DemoLogf("Please enter value to be generated [0x%.2X] ", nVal);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nVal = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
      
      DemoLogf("Please enter highway [0x%.2X] ", nHighway);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nHighway = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
      
      DemoLogf("Please enter TS [0x%.2X] ", nTs);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nTs = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
      
      nCmd = (nHighway << 3*8) | (nTs << 2*8) | (nVal << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hPhysDevs.hTss, nCmd);
      
    }
    break;

  
  case 'n':
    {
      OTX_UINT32 nValue;
      char sValue[80];
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
      OsaConsoleGetStr(sValue);
      if (sValue[0] == '\0') {
        break;
      }
      nValue = strtol(sValue, NULL, 0);
      
      // before setting the new trigger value, the burst device must not be enabled
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvEnable(g_hPhysDevs.hBurst,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      DemoLogf("New queue notification value is %d\n", nValue);
      
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[A|J]      : Board Type used for the demo, where\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;
  envP->coDsp              = 1;
  envP->nCurrentDsp        = 0;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if (OsaToUpper(*(argv[i] + 2)) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        }

      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  if (g_sBurstDevs.mBurstSender.pTxFile) {
    fclose(g_sBurstDevs.mBurstSender.pTxFile);
    g_sBurstDevs.mBurstSender.pTxFile = NULL;
  }
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
    g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
  }
  
  if (g_pDemoLogFile) {
    fclose(g_pDemoLogFile);
  }
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("A66BurstEx, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
//    OsaConsoleExit();
//    exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n");
  {
    g_bStatsThreadActive = OTX_FALSE;
    g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}
